/*
 * Extended Attribute support for rsync.
 * Written by Jay Fenlason, vaguely based on the ACLs patch.
 *
 * Copyright (C) 2004 Red Hat, Inc.
 * Copyright (C) 2006 Wayne Davison
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "rsync.h"
#include "lib/sysxattr.h"

#ifdef SUPPORT_XATTRS

extern int dry_run;
extern int am_root;
extern int read_only;
extern int list_only;
extern unsigned int file_struct_len;

#define RSYNC_XAL_INITIAL 5
#define RSYNC_XAL_LIST_INITIAL 100

#define HAS_PREFIX(str, prfx) (*(str) == *(prfx) \
			    && strncmp(str, prfx, sizeof (prfx) - 1) == 0)

#define USER_PREFIX "user."
#define UPRE_LEN ((int)sizeof USER_PREFIX - 1)
#define SYSTEM_PREFIX "system."
#define SPRE_LEN ((int)sizeof SYSTEM_PREFIX - 1)

#ifdef HAVE_LINUX_XATTRS
#define RPRE_LEN 0
#else
#define RSYNC_PREFIX "rsync."
#define RPRE_LEN ((int)sizeof RSYNC_PREFIX - 1)
#endif

typedef struct {
	char *datum, *name;
	size_t datum_len, name_len;
} rsync_xa;

static size_t namebuf_len = 0;
static char *namebuf = NULL;

static item_list empty_xattr = EMPTY_ITEM_LIST;
static item_list rsync_xal_l = EMPTY_ITEM_LIST;

/* ------------------------------------------------------------------------- */

static void rsync_xal_free(item_list *xalp)
{
	size_t i;
	rsync_xa *rxas = xalp->items;

	for (i = 0; i < xalp->count; i++) {
		free(rxas[i].datum);
		/*free(rxas[i].name);*/
	}
	xalp->count = 0;
}

void free_xattr(statx *sxp)
{
	rsync_xal_free(sxp->xattr);
	free(sxp->xattr);
	sxp->xattr = NULL;
}

static int rsync_xal_compare_names(const void *x1, const void *x2)
{
	const rsync_xa *xa1 = x1;
	const rsync_xa *xa2 = x2;
	return strcmp(xa1->name, xa2->name);
}

static int rsync_xal_get(const char *fname, item_list *xalp)
{
	ssize_t list_len, name_len, datum_len;
	char *name, *ptr;

	if (!namebuf) {
		namebuf_len = 1024;
		namebuf = new_array(char, namebuf_len);
		if (!namebuf)
			out_of_memory("rsync_xal_get");
	}

	/* The length returned includes all the '\0' terminators. */
	list_len = sys_llistxattr(fname, namebuf, namebuf_len);
	if (list_len > (ssize_t)namebuf_len) {
		list_len = -1;
		errno = ERANGE;
	}
	if (list_len < 0) {
		if (errno == ENOTSUP)
			return 0;
		if (errno == ERANGE) {
			list_len = sys_llistxattr(fname, NULL, 0);
			if (list_len < 0) {
				rsyserr(FERROR, errno,
					"rsync_xal_get: llistxattr(\"%s\",0) failed",
					fname);
				return -1;
			}
			namebuf = realloc_array(namebuf, char, list_len + 1024);
			if (!namebuf)
				out_of_memory("rsync_xal_get");
			namebuf_len = list_len + 1024;
			list_len = sys_llistxattr(fname, namebuf, namebuf_len);
			if (list_len < 0) {
				rsyserr(FERROR, errno,
					"rsync_xal_get: llistxattr(\"%s\",%ld) failed",
					fname, (long)namebuf_len);
				return -1;
			}
		} else {
			rsyserr(FERROR, errno,
				"rsync_xal_get: llistxattr(\"%s\",%ld) failed",
				fname, (long)namebuf_len);
			return -1;
		}
	}

	for (name = namebuf; list_len > 0; name += name_len) {
		rsync_xa *rxas;

		name_len = strlen(name) + 1;
		list_len -= name_len;

#ifdef HAVE_LINUX_XATTRS
		/* We don't send the system namespace. */
		if (HAS_PREFIX(name, SYSTEM_PREFIX))
			continue;
#endif

		datum_len = sys_lgetxattr(fname, name, NULL, 0);
		if (datum_len < 0) {
			if (errno == ENOTSUP)
				return -1;
			rsyserr(FERROR, errno,
			    "rsync_xal_get: lgetxattr(\"%s\",\"%s\",0) failed",
			    fname, name);
			return -1;
		}
		ptr = new_array(char, name_len + datum_len);
		if (!ptr)
			out_of_memory("rsync_xal_get");
		if (datum_len) {
			ssize_t len = sys_lgetxattr(fname, name, ptr, datum_len);
			if (len != datum_len) {
				if (len < 0) {
					rsyserr(FERROR, errno,
					    "rsync_xal_get: lgetxattr(\"%s\",\"%s\",%ld)"
					    " failed", fname, name, (long)datum_len);
				} else {
					rprintf(FERROR,
					    "rsync_xal_get: lgetxattr(\"%s\",\"%s\",%ld)"
					    " returned %ld\n", fname, name,
					    (long)datum_len, (long)len);
				}
				free(ptr);
				return -1;
			}
		}
		rxas = EXPAND_ITEM_LIST(xalp, rsync_xa, RSYNC_XAL_INITIAL);
		rxas->name = ptr + datum_len;
		rxas->datum = ptr;
		rxas->name_len = name_len;
		rxas->datum_len = datum_len;
		memcpy(rxas->name, name, name_len);
	}
	if (xalp->count > 1)
		qsort(xalp->items, xalp->count, sizeof (rsync_xa), rsync_xal_compare_names);
	return 0;
}

/* Read the xattr(s) for this filename. */
int get_xattr(const char *fname, statx *sxp)
{
	sxp->xattr = new(item_list);
	*sxp->xattr = empty_xattr;
	if (rsync_xal_get(fname, sxp->xattr) < 0) {
		free_xattr(sxp);
		return -1;
	}
	return 0;
}

static int find_matching_xattr(item_list *xalp)
{
	size_t i, j;
	item_list *lst = rsync_xal_l.items;

	for (i = 0; i < rsync_xal_l.count; i++) {
		rsync_xa *rxas1 = lst[i].items;
		rsync_xa *rxas2 = xalp->items;

		/* Wrong number of elements? */
		if (lst[i].count != xalp->count)
			continue;
		/* any elements different? */
		for (j = 0; j < xalp->count; j++) {
			if (rxas1[j].name_len != rxas2[j].name_len
			 || rxas1[j].datum_len != rxas2[j].datum_len
			 || strcmp(rxas1[j].name, rxas2[j].name)
			 || memcmp(rxas1[j].datum, rxas2[j].datum, rxas2[j].datum_len))
				break;
		}
		/* no differences found.  This is The One! */
		if (j == xalp->count)
			return i;
	}

	return -1;
}

/* Store *xalp on the end of rsync_xal_l */
static void rsync_xal_store(item_list *xalp)
{
	item_list *new_lst = EXPAND_ITEM_LIST(&rsync_xal_l, item_list, RSYNC_XAL_LIST_INITIAL);
	/* Since the following call starts a new list, we know it will hold the
	 * entire initial-count, not just enough space for one new item. */
	*new_lst = empty_xattr;
	(void)EXPAND_ITEM_LIST(new_lst, rsync_xa, xalp->count);
	memcpy(new_lst->items, xalp->items, xalp->count * sizeof (rsync_xa));
	new_lst->count = xalp->count;
	xalp->count = 0;
}

/* Send the make_xattr()-generated xattr list for this flist entry. */
void send_xattr(statx *sxp, int f)
{
	int ndx = find_matching_xattr(sxp->xattr);
	if (ndx != -1) {
		write_byte(f, 'x');
		write_int(f, ndx);
		rsync_xal_free(sxp->xattr);
	} else {
		rsync_xa *rxa;
		int count = sxp->xattr->count;
		write_byte(f, 'X');
		write_int(f, count);
		for (rxa = sxp->xattr->items; count--; rxa++) {
#ifdef HAVE_LINUX_XATTRS
			write_int(f, rxa->name_len);
			write_int(f, rxa->datum_len);
			write_buf(f, rxa->name, rxa->name_len);
#else
			/* We strip the rsync prefix from disguised namespaces
			 * and put everything else in the user namespace. */
			if (HAS_PREFIX(rxa->name, RSYNC_PREFIX)
			 && rxa->name[RPRE_LEN] != '%') {
				write_int(f, rxa->name_len - RPRE_LEN);
				write_int(f, rxa->datum_len);
				write_buf(f, rxa->name + RPRE_LEN, rxa->name_len - RPRE_LEN);
			} else {
				write_int(f, rxa->name_len + UPRE_LEN);
				write_int(f, rxa->datum_len);
				write_buf(f, USER_PREFIX, UPRE_LEN);
				write_buf(f, rxa->name, rxa->name_len);
			}
#endif
			write_buf(f, rxa->datum, rxa->datum_len);
		}
		rsync_xal_store(sxp->xattr); /* adds item to rsync_xal_l */
	}
	free_xattr(sxp);
}

/* ------------------------------------------------------------------------- */

/* receive and build the rsync_xattr_lists */
void receive_xattr(struct file_struct *file, int f)
{
	static item_list temp_xattr = EMPTY_ITEM_LIST;
	char *ndx_ptr = (char*)file + file_struct_len;
	int ndx, tag = read_byte(f);

	if (tag == 'X') {
		int i, count = read_int(f);
		for (i = 0; i < count; i++) {
			char *ptr, *name;
			rsync_xa *rxa;
			size_t name_len = read_int(f);
			size_t datum_len = read_int(f);
#ifdef HAVE_LINUX_XATTRS
			size_t extra_len = 0;
#else
			size_t extra_len = am_root ? RPRE_LEN : 0;
			if (datum_len + extra_len < datum_len)
				out_of_memory("receive_xattr"); /* overflow */
#endif
			if (name_len + datum_len + extra_len < name_len)
				out_of_memory("receive_xattr"); /* overflow */
			ptr = new_array(char, name_len + datum_len + extra_len);
			if (!ptr)
				out_of_memory("receive_xattr");
			name = ptr + datum_len + extra_len;
			read_buf(f, name, name_len);
			read_buf(f, ptr, datum_len);
#ifdef HAVE_LINUX_XATTRS
			/* Non-root can only save the user namespace. */
			if (!am_root && !HAS_PREFIX(name, USER_PREFIX)) {
				free(ptr);
				continue;
			}
#else
			/* This OS only has a user namespace, so we either
			 * strip the user prefix, or we put a non-user
			 * namespace inside our rsync hierarchy. */
			if (HAS_PREFIX(name, USER_PREFIX)) {
				name += UPRE_LEN;
				name_len -= UPRE_LEN;
			} else if (am_root) {
				name -= RPRE_LEN;
				name_len += RPRE_LEN;
				memcpy(name, RSYNC_PREFIX, RPRE_LEN);
			} else {
				free(ptr);
				continue;
			}
#endif
			rxa = EXPAND_ITEM_LIST(&temp_xattr, rsync_xa, count);
			rxa->name = name;
			rxa->datum = ptr;
			rxa->name_len = name_len;
			rxa->datum_len = datum_len;
		}
		ndx = rsync_xal_l.count; /* pre-incremented count */
		rsync_xal_store(&temp_xattr); /* adds item to rsync_xal_l */
	} else if (tag == 'x') {
		ndx = read_int(f);
		if (ndx < 0 || (size_t)ndx >= rsync_xal_l.count) {
			rprintf(FERROR, "receive_xattr: xa index %d out of"
				" range for %s\n", ndx, f_name(file, NULL));
			exit_cleanup(RERR_STREAMIO);
		}
	} else {
		rprintf(FERROR, "receive_xattr: unknown extended attribute"
			" type tag (%c) for %s\n", tag, f_name(file, NULL));
		exit_cleanup(RERR_STREAMIO);
	}

	SIVAL(ndx_ptr, 0, ndx);
}

/* Turn the xattr data in statx into cached xattr data, setting the index
 * values in the file struct. */
void cache_xattr(struct file_struct *file, statx *sxp)
{
	char *ndx_ptr = (char*)file + file_struct_len;
	int ndx;

	if (!sxp->xattr)
		return;

	ndx = find_matching_xattr(sxp->xattr);
	if (ndx == -1)
		rsync_xal_store(sxp->xattr); /* adds item to rsync_xal_l */
	free_xattr(sxp);

	SIVAL(ndx_ptr, 0, ndx);
}

static int rsync_xal_set(const char *fname, item_list *xalp)
{
	rsync_xa *rxas = xalp->items;
	size_t i;
	int ret = 0;

	for (i = 0; i < xalp->count; i++) {
		int status = sys_lsetxattr(fname, rxas[i].name, rxas[i].datum, rxas[i].datum_len);
		if (status < 0) {
			rsyserr(FERROR, errno,
				"rsync_xal_set: lsetxattr(\"%s\",\"%s\") failed",
				fname, rxas[i].name);
			ret = -1;
		}
	}
	return ret;
}

/* Set extended attributes on indicated filename. */
int set_xattr(const char *fname, const struct file_struct *file, UNUSED(statx *sxp))
{
	int ndx;
	char *ndx_ptr = (char*)file + file_struct_len;
	item_list *lst = rsync_xal_l.items;

	if (dry_run)
		return 1; /* FIXME: --dry-run needs to compute this value */

	if (read_only || list_only) {
		errno = EROFS;
		return -1;
	}

	ndx = IVAL(ndx_ptr, 0);
	return rsync_xal_set(fname, lst + ndx); /* TODO:  This needs to return 1 if no xattrs changed! */
}

#endif /* SUPPORT_XATTRS */
